"""
Namen nennenswerter Punkte, die als Fixierpositionen für eine Grafik verwendet werden können
"""

from __future__ import annotations

from pytamaro.point_names import (top_left, top_center, top_right,
                                  center_left, center, center_right,
                                  bottom_left, bottom_center, bottom_right)
from pytamaro.de.point import Punkt


mitte: Punkt = center
"""Der Mittelpunkt des Begrenzungsrahmens

:meta hide-value:
"""


oben_mitte: Punkt = top_center
"""Der Mittelpunkt der oberen Kante des Begrenzungsrahmens

:meta hide-value:
"""

unten_mitte: Punkt = bottom_center
"""Der Mittelpunkt der unteren Kante des Begrenzungsrahmens

:meta hide-value:
"""

mitte_links: Punkt = center_left
"""Der Mittelpunkt der linken Kante des Begrenzungsrahmens

:meta hide-value:
"""


mitte_rechts: Punkt = center_right
"""Der Mittelpunkt der rechten Kante des Begrenzungsrahmens

:meta hide-value:
"""


oben_links: Punkt = top_left
"""Die obere linke Ecke des Begrenzungsrahmens

:meta hide-value:
"""

oben_rechts: Punkt = top_right
"""Die obere rechte Ecke des Begrenzungsrahmens

:meta hide-value:
"""

unten_links: Punkt = bottom_left
"""Die untere linke Ecke des Begrenzungsrahmens

:meta hide-value:
"""

unten_rechts: Punkt = bottom_right
"""Die untere rechte Ecke des Begrenzungsrahmens

:meta hide-value:
"""



"""
General utils.
"""

import sys


def export(fn):  # pylint: disable=invalid-name
    """
    Use a decorator to avoid retyping function/class names.
    """
    mod = sys.modules[fn.__module__]
    if hasattr(mod, '__all__'):
        name = fn.__name__
        all_ = mod.__all__
        if name not in all_:
            all_.append(name)
    else:
        setattr(mod, '__all__', [fn.__name__])
    return fn


def is_notebook() -> bool:
    """
    Checks whether we are running inside a Jupyter notebook.
    Adapted from
    https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook/24937408

    :returns: True if running inside a notebook, False otherwise
    """
    try:
        # pylint: disable=import-outside-toplevel
        from IPython import get_ipython  # type: ignore[import]
        if 'IPKernelApp' not in get_ipython().config:
            return False
    except ImportError:
        return False
    except AttributeError:
        return False
    return True



"""
Names of notable colors because they are at the vertices of the RGB cube,
plus the fully-transparent one.
"""

from pytamaro.color_functions import rgb_color

black = rgb_color(0, 0, 0)
"""Black color

:meta hide-value:
"""

red = rgb_color(255, 0, 0)
"""Red color

:meta hide-value:
"""

green = rgb_color(0, 255, 0)
"""Green color

:meta hide-value:
"""

blue = rgb_color(0, 0, 255)
"""Blue color

:meta hide-value:
"""

yellow = rgb_color(255, 255, 0)
"""Yellow color

:meta hide-value:
"""

magenta = rgb_color(255, 0, 255)
"""Magenta color

:meta hide-value:
"""

cyan = rgb_color(0, 255, 255)
"""Cyan color

:meta hide-value:
"""

white = rgb_color(255, 255, 255)
"""White color

:meta hide-value:
"""

transparent = rgb_color(0, 0, 0, 0)
"""Fully-transparent color

:meta hide-value:
"""



"""Localization utilities to translate error messages"""
# pylint: disable=line-too-long

import sys

translations = {
    # Error messages
    "INVALID_LENGTH": {
        "en": "Invalid length for parameter {}",
        "it": "Lunghezza non valida per il parametro {}",
        "de": "Ungültige Länge für den Parameter {}",
    },
    "INVALID_TYPE": {
        "en": "Invalid type for parameter {}: expected {}, got {}",
        "it": "Tipo non valido per il parametro {}: previsto {}, ottenuto {}",
        "de": "Ungültiger Typ für den Parameter {}: erwartet {}, erhalten {}",
    },
    "INVALID_RANGE": {
        "en": "Value for parameter {} out of the allowed range [{}, {}]",
        "it": "Valore per il parametro {} fuori dall'intervallo consentito [{}, {}]",
        "de": "Wert für den Parameter {} außerhalb des zulässigen Bereichs [{}, {}]",
    },
    "INVALID_FILENAME_EXTENSION": {
        "en": "Invalid or missing extension for filename: only .png and .svg are supported",
        "it": "Estensione non valida o mancante per il nome del file: sono supportati solo .png e .svg",
        "de": "Ungültige oder fehlende Dateinamenerweiterung: Nur .png und .svg werden unterstützt",
    },
    "INVALID_FILENAME_GIF": {
        "en": "Invalid or missing extension for a GIF file: needs to end with `.gif`",
        "it": "Estensione non valida o mancante per un file GIF: deve terminare con `.gif`",
        "de": "Ungültige oder fehlende Dateinamenerweiterung für eine GIF-Datei: muss mit `.gif` enden",
    },
    "EMPTY_GRAPHICS_LIST": {
        "en": "The list of graphics cannot be empty",
        "it": "La lista delle grafiche non può essere vuota",
        "de": "Die Liste der Grafiken darf nicht leer sein",
    },
    "EMPTY_AREA_OUTPUT": {
        "en": "Cannot show/save a graphic of size {} as it has no area",
        "it": "Impossibile mostrare/salvare una grafica di dimensione {} poiché non ha area",
        "de": "Kann eine Grafik der Größe {} nicht anzeigen/speichern, da sie keine Fläche hat",
    },
    # Parameter names
    "width": {
        "en": "width",
        "it": "larghezza",
        "de": "breite",
    },
    "height": {
        "en": "height",
        "it": "altezza",
        "de": "hoehe",
    },
    "radius": {
        "en": "radius",
        "it": "raggio",
        "de": "radius",
    },
    "red": {
        "en": "red",
        "it": "rosso",
        "de": "rot",
    },
    "green": {
        "en": "green",
        "it": "verde",
        "de": "gruen",
    },
    "blue": {
        "en": "blue",
        "it": "blu",
        "de": "blau",
    },
    "opacity": {
        "en": "opacity",
        "it": "opacita",
        "de": "opazitaet",
    },
    "hue": {
        "en": "hue",
        "it": "tonalita",
        "de": "farbton",
    },
    "saturation": {
        "en": "saturation",
        "it": "saturazione",
        "de": "saettigung",
    },
    "value": {
        "en": "value",
        "it": "valore",
        "de": "hellwert",
    },
    "lightness": {
        "en": "lightness",
        "it": "luce",
        "de": "helligkeit",
    },
    "angle": {
        "en": "angle",
        "it": "angolo",
        "de": "winkel",
    },
    "side1": {
        "en": "side1",
        "it": "lato1",
        "de": "seite1",
    },
    "side2": {
        "en": "side2",
        "it": "lato2",
        "de": "seite2",
    },
    "content": {
        "en": "content",
        "it": "contenuto",
        "de": "inhalt",
    },
    "foreground_graphic": {
        "en": "foreground_graphic",
        "it": "grafica_primopiano",
        "de": "vordere_grafik",
    },
    "background_graphic": {
        "en": "background_graphic",
        "it": "grafica_secondopiano",
        "de": "hintere_grafik",
    },
    "left_graphic": {
        "en": "left_graphic",
        "it": "grafica_sinistra",
        "de": "linke_grafik",
    },
    "right_graphic": {
        "en": "right_graphic",
        "it": "grafica_destra",
        "de": "rechte_grafik",
    },
    "top_graphic": {
        "en": "top_graphic",
        "it": "grafica_alto",
        "de": "obere_grafik",
    },
    "bottom_graphic": {
        "en": "bottom_graphic",
        "it": "grafica_basso",
        "de": "untere_grafik",
    },
    "filename": {
        "en": "filename",
        "it": "nome_file",
        "de": "datei_name",
    },
    "graphics": {
        "en": "graphics",
        "it": "grafiche",
        "de": "grafiken",
    },
    "graphic": {
        "en": "graphic",
        "it": "grafica",
        "de": "grafiken",
    },
    "color": {
        "en": "color",
        "it": "colore",
        "de": "farbe",
    },
    "point": {
        "en": "point",
        "it": "punto",
        "de": "punkt",
    },
    "font": {
        "en": "font",
        "it": "font",
        "de": "schriftart",
    },
    "points": {
        "en": "points",
        "it": "punti",
        "de": "punkte",
    },
    # Types
    "Graphic": {
        "en": "Graphic",
        "it": "Grafica",
        "de": "Grafik",
    },
    "Color": {
        "en": "Color",
        "it": "Colore",
        "de": "Farbe",
    },
    "Point": {
        "en": "Point",
        "it": "Punto",
        "de": "Punkt",
    },
}


def translate(key: str, *args) -> str:
    """Looks up an error message by `key` and returns the localized version.

    :param key: error message key
    :param args: optional arguments to be formatted into the localized message
    :returns: locale-specific descriptive error message
    """
    lang = sys.modules["pytamaro"].LANGUAGE
    if key not in translations:
        return key
    localized = translations[key][lang]
    if len(localized) == 0:
        localized = translations[key]["en"]
    return localized.format(*args)



"""
Der Typ `Farbe`, Funktionen, die Farben erzeugen
und Konstanten für wichtige Farben.
"""

from __future__ import annotations

from pytamaro.color import Color
from pytamaro.color_functions import rgb_color, hsl_color, hsv_color

Farbe = Color
"""
Repräsentiert eine Farbe.
Eine Farbe hat auch eine gewisse Opazität,
von komplett durchsichtig (wie die Farbe `transparent`),
bis komplett undurchsichtig (wie die Farbe `rot`).
"""


def rgb_farbe(rot: int, gruen: int, blau: int, opazitaet: float = 1.0) -> Farbe:
    """
    Erzeugt eine Farbe mit den gegebenen Anteilen Rot (R), Grün (G) und Blau (B)
    und der gegebenen Opazität (Undurchsichtigkeit, alpha, A).

    .. figure:: https://upload.wikimedia.org/wikipedia/commons/thumb/b/b4/RGBCube_a.svg/524px-RGBCube_a.svg.png
       :height: 120px
       :align: center

       `RGB Würfel (SharkD via Wikimedia Commons) <https://commons.wikimedia.org/wiki/File:RGBCube_a.svg>`_

    :param rot: der rote Farbanteil [0-255]
    :param gruen: der grüne Farbanteil [0-255]
    :param blau: der blaue Farbanteil [0-255]
    :param opazitaet: die Undurchsichtigkeit (alpha),
           wobei 0 komplett durchsichtig und 1 komplett undurchsichtig
           entspricht. Standardmäßig sind alle Farben vollständig undurchsichtig.
    :returns: eine Farbe mit den gegebenen RGBA-Komponenten
    """
    return rgb_color(rot, gruen, blau, opazitaet)


def hsv_farbe(farbton: float, saettigung: float, hellwert: float, opazitaet: float = 1.0) -> Farbe:
    """
    Erzeugt eine Farbe mit dem gegebenen Farbton (H), der Sättigung (S),
    dem Hellwert (V) und der Opazität (Undurchsichtigkeit, alpha, A).

    .. figure:: https://upload.wikimedia.org/wikipedia/commons/4/4e/HSV_color_solid_cylinder.png
       :height: 120px
       :align: center

       `HSV Zylinder (SharkD via Wikimedia Commons) <https://commons.wikimedia.org/wiki/File:HSV_color_solid_cylinder.png>`_

    :param farbton: der Farbton (hue) [0-360] als Farbwinkel, in Grad,
           auf dem Farbkreis (0 für Rot, 120 für Grün, 240 für Blau)
    :param saettigung: Farbsättigung (saturation) [0-1]
           (0 = Grau, 0.5 = wenig gesättigte Farbe, 1 = gesättigte, reine Farbe)
    :param hellwert: Hellwert (value) der Farbe [0-1]
           (0 = dunkel, 1 = hell)
    :param opazitaet: die Undurchsichtigkeit (alpha),
           wobei 0 komplett durchsichtig und 1 komplett undurchsichtig
           entspricht. Standardmäßig sind alle Farben vollständig undurchsichtig.
    :returns: eine Farbe mit den gegebenen HSVA-Komponenten
    """
    return hsv_color(farbton, saettigung, hellwert, opazitaet)


def hsl_farbe(farbton: float, saettigung: float, helligkeit: float,
              opazitaet: float = 1.0) -> Farbe:
    """
    Erzeugt eine Farbe mit dem gegebenen Farbton (H), der Sättigung (S),
    dem Helligkeit (L) und der Opazität (Undurchsichtigkeit, alpha, A).

    .. figure:: https://upload.wikimedia.org/wikipedia/commons/3/35/HSL_color_solid_cylinder.png
       :height: 120px
       :align: center

       `HSL Zylinder: SharkD via Wikimedia Commons <https://commons.wikimedia.org/wiki/File:HSL_color_solid_cylinder.png>`_

    :param farbton: der Farbton (hue) [0-360] als Farbwinkel, in Grad,
           auf dem Farbkreis (0 für Rot, 120 für Grün, 240 für Blau)
    :param saettigung: Farbsättigung (saturation) [0-1]
           (0 = Grau, 0.5 = wenig gesättigte Farbe, 1 = gesättigte, reine Farbe)
    :param helligkeit: der Anteil Schwarz oder Weiss [0-1].
           (0 = Schwarz, 0.5 = weder abgedunkelt noch aufgehellt, 1 = Weiss)
    :param opazitaet: die Undurchsichtigkeit (alpha),
           wobei 0 komplett durchsichtig und 1 komplett undurchsichtig
           entspricht. Standardmäßig sind alle Farben vollständig undurchsichtig.
    :returns: eine Farbe mit den gegebenen HSLA-Komponenten
    """
    return hsl_color(farbton, saettigung, helligkeit, opazitaet)



"""
Functions for output (show or save) of graphics.
"""

import base64
import io
import os
import re
from pathlib import Path
import subprocess
import sys
from tempfile import NamedTemporaryFile
from typing import List

from PIL import Image as PILImageMod
from PIL.Image import Image as PILImage
from skia import Canvas, FILEWStream, Image, Rect, Surface, SVGCanvas, kPNG

from pytamaro.checks import check_graphic, check_type
from pytamaro.debug import add_debug_info
from pytamaro.graphic import Graphic
from pytamaro.localization import translate
from pytamaro.utils import export, is_notebook


def _warning_no_area(graphic: Graphic):
    """
    Emits a warning indicating that the graphic cannot be shown or saved
    because it has no area.

    :param graphic: graphic that cannot be shown or saved
    """
    size = graphic.bounds().round()
    # pylint: disable-next=line-too-long
    print(translate("EMPTY_AREA_OUTPUT", f"{size.width()}x{size.height()}"))


def _draw_to_canvas(canvas: Canvas, graphic: Graphic):
    """
    Draws a graphic to a canvas, correcting for the top-left position.

    :param canvas: canvas onto which to draw
    :param graphic: graphic to be drawn
    """
    bounds = graphic.bounds()
    canvas.translate(-bounds.left(), -bounds.top())
    graphic.draw(canvas)


# pylint: disable-next=invalid-name
def _save_as_SVG(filename: str, graphic: Graphic):
    """
    Save a graphic to an SVG file.

    :param filename: name of the file to be created, ending in ".svg"
    :param graphic: graphic to be saved
    """
    size = graphic.size()
    stream = FILEWStream(filename)
    canvas = SVGCanvas.Make(Rect.MakeSize(size), stream)
    _draw_to_canvas(canvas, graphic)
    del canvas
    stream.flush()
    # Manually add shape-rendering="crispEdges" to the SVG file.
    # We don't use the XML parser from the standard library because,
    # among other aspects, it does not properly maintain the doctype.
    with open(filename, "r", encoding="utf-8") as file:
        content = file.read()
    # `svg` tag may be self-closing
    new_content = re.sub("<svg(.*?)(/?)>", r'<svg\1 shape-rendering="crispEdges"\2>', content)
    with open(filename, "w", encoding="utf-8") as file:
        file.write(new_content)


def graphic_to_image(graphic: Graphic) -> Image:
    """
    Renders a graphic into a Skia image.

    :param graphic: graphic to be rendered
    :returns: rendered graphic as a Skia image
    """
    int_size = graphic.size().toCeil()
    surface = Surface(int_size.width(), int_size.height())
    _draw_to_canvas(surface.getCanvas(), graphic)
    return surface.makeImageSnapshot()


def graphic_to_pillow_image(graphic: Graphic) -> PILImage:
    """
    Renders a graphic and converts it into a Pillow image.

    :param graphic: graphic to be rendered and converted
    :returns: rendered graphic as a Pillow image
    """
    with io.BytesIO(graphic_to_image(graphic).encodeToData()) as stream:
        pil_image = PILImageMod.open(stream)
        pil_image.load()  # Ensure to make a copy of buffer
        return pil_image


# pylint: disable-next=invalid-name
def _save_as_PNG(filename: str, graphic: Graphic):
    """
    Save a graphic to a PNG file.

    :param filename: name of the file to be created, ending in ".png"
    :param graphic: graphic to be saved
    """
    graphic_to_image(graphic).save(filename, kPNG)


@export
def show_graphic(graphic: Graphic, debug: bool = False):
    """
    Show a graphic. Graphics with no area cannot be shown.

    When `debug` is `True`, adorns the visualization with useful information
    for debugging: a red border around the bounding box and a yellowish cross
    around the pinning position.

    :param graphic: graphic to be shown
    :param debug: can be optionally set to `True` to overlay debugging
           information
    """
    check_graphic(graphic)
    if graphic.empty_area():
        _warning_no_area(graphic)
    else:
        to_show = add_debug_info(graphic) if debug else graphic
        pil_image = graphic_to_pillow_image(to_show)
        if is_notebook():
            # pylint: disable-next=undefined-variable
            display(pil_image)  # type: ignore[name-defined]
        elif "PYTAMARO_OUTPUT_DATA_URI" in os.environ:
            buffer = io.BytesIO()
            pil_image.save(buffer, format="PNG")
            b64_str = base64.b64encode(buffer.getvalue()).decode("utf-8")
            print(f"data:image/png;base64,{b64_str}", end="")
        else:
            pil_image.show()


@export
def save_graphic(filename: str, graphic: Graphic, debug: bool = False):
    """
    Save a graphic to a file.
    Two file formats are supported: PNG (raster graphics) and SVG (vector graphics).
    The extension of the filename (either ".png" or ".svg") determines the format.

    Graphics with no area cannot be saved in the PNG format.

    When `debug` is `True`, adorns the visualization with useful information
    for debugging: a red border around the bounding box and a yellowish cross
    around the pinning position.

    :param filename: name of the file to create (with the extension)
    :param graphic: graphic to be saved
    :param debug: can be optionally set to `True` to overlay debugging
           information
    """
    check_type(filename, str, "filename")
    check_graphic(graphic)
    to_show = add_debug_info(graphic) if debug else graphic
    extension = Path(filename).suffix
    if extension == ".png":
        if graphic.empty_area():
            _warning_no_area(graphic)
        else:
            _save_as_PNG(filename, to_show)
    elif extension == ".svg":
        _save_as_SVG(filename, to_show)
    else:
        raise ValueError(translate("INVALID_FILENAME_EXTENSION"))


@export
def save_animation(filename: str, graphics: List[Graphic], duration: int = 40, loop: bool = True):
    """
    Save a sequence of graphics as an animation (GIF).

    Graphics are sequentially reproduced (normally at 25 frames per second) in
    a loop (unless specificied otherwise).

    :param filename: name of the file to create, including the extension '.gif'
    :param graphics: list of graphics to be saved as an animation
    :param duration: duration in milliseconds for each frame
           (defaults to 40 milliseconds, which leads to 25 frames per second)
    :param loop: whether the GIF should loop indefinitely (defaults to true)
    """
    check_type(filename, str, "filename")
    if Path(filename).suffix != ".gif":
        raise ValueError(translate("INVALID_FILENAME_GIF"))
    check_type(graphics, list, "graphics")
    if len(graphics) == 0:
        raise ValueError(translate("EMPTY_GRAPHICS_LIST"))
    pil_images = list(map(graphic_to_pillow_image, graphics))
    pil_images[0].save(
        filename,
        save_all=True,
        append_images=pil_images[1:],
        duration=duration,
        loop=0 if loop else 1,  # loop 0 means "indefinitely", 1 means "once"
    )


@export
def show_animation(graphics: List[Graphic], duration: int = 40, loop: bool = True):
    """
    Show a sequence of graphics as an animation (GIF).

    Graphics are sequentially reproduced (normally at 25 frames per second) in
    a loop (unless specificied otherwise).

    :param graphics: list of graphics to be shown as an animation
    :param duration: duration in milliseconds for each frame
           (defaults to 40 milliseconds, which leads to 25 frames per second)
    :param loop: whether the animation should loop indefinitely (defaults to true)
    """
    with NamedTemporaryFile(suffix=".gif", delete=False) as file:
        save_animation(file.name, graphics, duration, loop)
        if is_notebook():
            # pylint: disable-next=import-outside-toplevel, import-error
            from IPython.display import Image as IPythonImage  # type: ignore[import]
            with open(file.name, "rb") as stream:
                # pylint: disable-next=undefined-variable
                display(IPythonImage(stream.read()))  # type: ignore[name-defined]
        elif "PYTAMARO_OUTPUT_DATA_URI" in os.environ:
            with open(file.name, "rb") as stream:
                b64_str = base64.b64encode(stream.read()).decode("utf-8")
                print(f"data:image/gif;base64,{b64_str}", end="")
        elif sys.platform == "win32":
            os.startfile(file.name)
        elif sys.platform == "darwin":
            subprocess.call(["open", "-a", "Safari", file.name])
        else:
            subprocess.call(["xdg-open", file.name])



"""
Die folgenden nennenswerten Farben sind bereits definiert:
Farben an den Ecken des RGB Würfels, komplett durchsichtige Farbe.
"""

from __future__ import annotations

from pytamaro.color_names import (black, blue, cyan, green, magenta, red,
                                  transparent, white, yellow)
from pytamaro.de.color import Farbe

schwarz: Farbe = black
"""
Die Farbe Schwarz.

:meta hide-value:
"""

rot: Farbe = red
"""
Die Farbe Rot.

:meta hide-value:
"""

gruen: Farbe = green
"""
Die Farbe Grün.

:meta hide-value:
"""

blau: Farbe = blue
"""
Die Farbe Blau.

:meta hide-value:
"""

gelb: Farbe = yellow
"""
Die Farbe Gelb.

:meta hide-value:
"""

magenta: Farbe = magenta  # pylint: disable=self-assigning-variable
"""
Die Farbe Magenta.

:meta hide-value:
"""

cyan: Farbe = cyan  # pylint: disable=self-assigning-variable
"""
Die Farbe Cyan.

:meta hide-value:
"""

weiss: Farbe = white
"""
Die Farbe Weiss.

:meta hide-value:
"""

transparent: Farbe = transparent  # pylint: disable=self-assigning-variable
"""
Die komplett durchsichtige Farbe.

:meta hide-value:
"""



"""
Funzioni per operazioni con grafiche (principlamente per combinarle).
"""

from __future__ import annotations

from pytamaro.it.graphic import Grafica
from pytamaro.it.point import Punto
from pytamaro.operations import (above, beside, compose, graphic_height,
                                 graphic_width, overlay, pin, rotate)


def larghezza_grafica(grafica: Grafica) -> int:
    """
    Ritorna la larghezza di una grafica.

    :param grafica: grafica di cui calcolare la larghezza
    :returns: larghezza della grafica
    """
    return graphic_width(grafica)


def altezza_grafica(grafica: Grafica) -> int:
    """
    Ritorna l'altezza di una grafica.

    :param grafica: grafica di cui calcolare l'altezza
    :returns: altezza della grafica
    """
    return graphic_height(grafica)


def componi(grafica_primopiano: Grafica, grafica_secondopiano: Grafica) \
        -> Grafica:
    """
    Crea una nuova grafica componendo le due grafiche fornite.
    La prima grafica viene tenuta in primo piano, la seconda sullo sfondo.
    Le grafiche vengono allineate superimponendo le loro posizioni di fissaggio.

    La posizione di fissaggio usata per comporre diventa la posizione di
    fissaggio della grafica risultante.

    :param grafica_primopiano: grafica in primo piano
    :param grafica_secondopiano: grafica sullo sfondo
    :returns: la grafica risultante composta
    """
    return compose(grafica_primopiano, grafica_secondopiano)


def fissa(punto: Punto, grafica: Grafica) -> Grafica:
    """
    Crea una nuova grafica che corrisponde alla grafica fornita,
    con una nuova posizione di fissaggio.

    Ogni grafica è racchiusa in un rettangolo di delimitazione (bounding box).
    Ci sono 9 punti notevoli, corrispondenti ai quattro angoli di questo rettangolo,
    ai punti centrali dei quattro lati e al centro del rettangolo.
    Ci si può riferire a questi punti usando i nomi `alto_sinistra`, `alto_destra`,
    `basso_sinistra`, `basso_destra`, `alto_centro`, `centro_destra`, `basso_centro`,
    `centro_sinistra` e `centro`.

    :param punto: il punto indicante la nuova posizione di fissaggio
    :param grafica: grafica originale
    :returns: una nuova grafica con una posizione di fissaggio aggiornata
    """
    return pin(punto, grafica)


def sovrapponi(grafica_primopiano: Grafica, grafica_secondopiano: Grafica) \
        -> Grafica:
    """
    Crea una nuova grafica sovrapponendo le due grafiche fornite,
    tenendo la prima in primo piano e la seconda sullo sfondo.
    Le due grafiche vengono sovrapposte sui loro centri.

    La posizione di fissaggio della grafica risultante è nel suo centro.

    :param grafica_primopiano: grafica in primo piano
    :param grafica_secondopiano: grafica sullo sfondo
    :returns: grafica risultante dalla sovrapposizione delle due fornite
    """
    return overlay(grafica_primopiano, grafica_secondopiano)


def accanto(grafica_sinistra: Grafica, grafica_destra: Grafica) -> Grafica:
    """
    Crea una nuova grafica affiancando orizzontalmente le due grafiche fornite.
    Le due grafiche vengono centrate verticalmente.

    La posizione di fissaggio della grafica risultante è nel suo centro.

    :param grafica_sinistra: grafica da posizionare a sinistra
    :param grafica_destra: grafica da posizionare a destra
    :returns: grafica risultante dall'affiancamento orizzontale delle due
              grafiche fornite
    """
    return beside(grafica_sinistra, grafica_destra)


def sopra(grafica_alto: Grafica, grafica_basso: Grafica) -> Grafica:
    """
    Crea una nuova grafica posizionando le due grafiche fornite una sopra l'altra.
    Le due grafiche vengono centrate orizzontalmente.

    La posizione di fissaggio della grafica risultante è nel suo centro.

    :param grafica_sopra: grafica da posizionare sopra
    :param grafica_sotto: grafica da posizionare sotto
    :returns: grafica risultante dall'affiancamento verticale delle due
              grafiche fornite
    """
    return above(grafica_alto, grafica_basso)


def ruota(angolo: float, grafica: Grafica) -> Grafica:
    """
    Crea una nuova grafica ruotando dell'angolo indicato, in senso antiorario,
    una grafica attorno alla sua posizione di fissaggio.
    Un angolo negativo corrisponde a una rotazione in senso orario.

    :param angolo: angolo di rotazione in senso antiorario, in gradi
    :param grafica: grafica da ruotare
    :returns: una nuova grafica, ruotata
    """
    return rotate(angolo, grafica)



"""
Type `Graphic`, that includes a graphic with a pinning position.
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass

from skia import (Canvas, Font, Matrix, Paint, Path, Point, Rect, Size,
                  Typeface)

from pytamaro.color import Color
from pytamaro.point import Point as PyTamaroPoint

# pylint: disable=super-init-not-called


@dataclass
class Graphic(ABC):
    """
    A graphic (image) with a position for pinning.

    The pinning position is used in the following operations:

    - rotation (to determine the center of rotation)
    - graphic composition (two graphics get composed aligning their pinning
      position).
    """
    pin_position: Point
    path: Path

    def set_pin_position(self, x: float, y: float):  # pylint: disable=invalid-name
        """
        Changes the pinning position of this graphic.

        :param x: x coordinate
        :param y: y coordinate
        """
        self.pin_position = Point(x, y)

    def size(self) -> Size:
        """
        Computes the size of this graphic (x and y axes spanning),
        using the bounds computed by bounds().

        :returns: graphic's size
        """
        bounds = self.bounds()
        return Size(bounds.width(), bounds.height())

    def bounds(self) -> Rect:
        """
        Computes the (tight) bounds for the path (outline) of this graphic.

        :returns: a rectangle that indicates the bounds of the graphic in the 2D
                  space
        """
        return self.path.computeTightBounds()

    @abstractmethod
    def draw(self, canvas: Canvas):
        """
        Draws the current graphic onto the provided canvas.

        :param canvas: canvas onto which to draw
        """

    def empty_area(self) -> bool:
        """
        Returns whether this graphic has an empty area (width or height 0) or
        not.

        :returns: True if the graphic has an empty area, False otherwise
        """
        return self.size().isEmpty()

    def __hash__(self) -> int:
        return hash(self._key())

    def _key(self):
        return ((self.pin_position.fX, self.pin_position.fY),
                self.path.serialize().bytes())


class Primitive(Graphic):
    """
    Represents a primitive graphic, which has a path and a uniform color.
    Geometric shapes and text are primitive graphics.
    """
    def __init__(self, path: Path, color: Color):
        self.path = path
        self.paint = Paint(color.color)
        bounds = self.path.computeTightBounds()
        self.set_pin_position(bounds.width() / 2, bounds.height() / 2)

    def draw(self, canvas: Canvas):
        canvas.drawPath(self.path, self.paint)

    def _key(self):
        return (super()._key(), self.paint.getHash())


class Empty(Graphic):
    """
    An empty graphic.
    """
    def __init__(self):
        super().__init__(Point(0, 0), Path())

    def draw(self, canvas: Canvas):
        pass


class Rectangle(Primitive):
    """
    A rectangle.
    """
    def __init__(self, width: float, height: float, color: Color):
        path = Path().addRect(Rect.MakeWH(width, height))
        super().__init__(path, color)


class Ellipse(Primitive):
    """
    An ellipse.
    """
    def __init__(self, width: float, height: float, color: Color):
        path = Path().addOval(Rect.MakeWH(width, height))
        super().__init__(path, color)


class CircularSector(Primitive):
    """
    A circular sector (with an angle between 0 and 360).
    Its pinning position is the center of the circle from which it is taken.
    """
    def __init__(self, radius: float, angle: float, color: Color):
        if angle == 360:
            path = Path.Circle(0, 0, radius)
        else:
            diameter = 2 * radius
            path = Path()
            path.moveTo(radius, radius)
            path.arcTo(Rect.MakeWH(diameter, diameter), 0, -angle, False)
            path.close()
        super().__init__(path, color)
        self.set_pin_position(radius, radius)


class Triangle(Primitive):
    """
    A triangle specified using two sides and the angle between them.
    The first side extends horizontally to the right.
    The second side is rotated counterclockwise by the specified angle.

    Its pinning position is the centroid of the triangle.
    """
    def __init__(self, side1: float, side2: float, angle: float, color: Color):
        third_point = Matrix.RotateDeg(-angle).mapXY(side2, 0)
        path = Path.Polygon([Point(0, 0), Point(side1, 0), third_point], isClosed=True)
        super().__init__(path, color)
        # The centroid is the average of the three vertices
        centroid = Point((side1 + third_point.x()) / 3, third_point.y() / 3)
        self.pin_position = centroid


class Text(Primitive):
    """
    Graphic containing text, using a given font with a given typographic size.
    Its pinning position is horizontally aligned on the left and vertically on
    the baseline of the text.
    """
    def __init__(self, text: str, font_name: str, size: float, color: Color):
        font = Font(Typeface(font_name), size)
        glyphs = font.textToGlyphs(text)
        offsets = font.getXPos(glyphs)
        text_path = Path()
        for glyph, x_offset in zip(glyphs, offsets):
            path = font.getPath(glyph)
            # Some glyphs (e.g., the space) have no outline
            if path is not None:
                path.offset(x_offset, 0)
                text_path.addPath(path)

        super().__init__(text_path, color)
        # Set the pinning position at baseline level (0) on the very left (which
        # might be slightly after 0, given that the bounding box is computed
        # very tightly around the glyphs, cutting some space before the first
        # one).
        self.set_pin_position(self.bounds().left(), 0)


class Compose(Graphic):
    """
    Represents the composition of two graphics, one in the foreground and the
    other in the background, joined on their pinning positions.
    """
    def __init__(self, foreground: Graphic, background: Graphic):
        self.foreground = foreground
        self.background = background
        fg_pin = self.foreground.pin_position
        bg_pin = self.background.pin_position
        self.set_pin_position(bg_pin.x(), bg_pin.y())
        self.path = Path(self.background.path)
        self.path.addPath(self.foreground.path,
                          bg_pin.x() - fg_pin.x(), bg_pin.y() - fg_pin.y())

    def draw(self, canvas: Canvas):
        canvas.save()
        self.background.draw(canvas)
        canvas.translate(self.background.pin_position.x() - self.foreground.pin_position.x(),
                         self.background.pin_position.y() - self.foreground.pin_position.y())
        self.foreground.draw(canvas)
        canvas.restore()


class Pin(Graphic):
    """
    Represents the pinning of a graphic in a certain position on its bounds.
    """
    def __init__(self, graphic: Graphic, pinning_point: PyTamaroPoint):
        self.graphic = graphic
        bounds = self.graphic.bounds()
        h_mapping = {
            -1.0: bounds.left(),
            0.0: bounds.centerX(),
            1.0: bounds.right()
        }
        v_mapping = {
            1.0: bounds.top(),
            0.0: bounds.centerY(),
            -1.0: bounds.bottom()
        }
        self.set_pin_position(
            h_mapping[pinning_point.x], v_mapping[pinning_point.y])
        self.path = Path(self.graphic.path)

    def draw(self, canvas: Canvas):
        self.graphic.draw(canvas)


class Rotate(Graphic):
    """
    Represents the rotation of a graphic by a certain angle clockwise.
    """
    def __init__(self, graphic: Graphic, deg: float):
        self.graphic = graphic
        self.rot_matrix = Matrix.RotateDeg(deg, self.graphic.pin_position)
        self.path = Path()
        self.graphic.path.transform(self.rot_matrix, self.path)  # updates self.path
        self.pin_position = graphic.pin_position

    def draw(self, canvas: Canvas):
        canvas.save()
        canvas.concat(self.rot_matrix)
        self.graphic.draw(canvas)
        canvas.restore()



"""
Der Typ `Grafik`, kombiniert eine Grafik mit einer Fixierposition.
"""

from __future__ import annotations

from pytamaro.graphic import Graphic

Grafik = Graphic
"""
Eine Grafik mit einer Fixierposition.

Die Fixierposition wird von den folgenden Operationen verwendet:

- `drehe` (das Rotationszentrum ist die Fixierposition)
- `kombiniere` (die zwei Grafiken werden so ausgerichtet,
  dass ihre Fixierpositionen übereinanderliegen).
"""



"""
Nomi di punti notevoli, che possono essere usati come posizioni di fissaggio per
una grafica.
"""

from __future__ import annotations

from pytamaro.point_names import (top_left, top_center, top_right,
                                  center_left, center, center_right,
                                  bottom_left, bottom_center, bottom_right)
from pytamaro.it.point import Punto


centro: Punto = center
"""Il punto centrale del rettangolo di delimitazione della grafica

:meta hide-value:
"""


alto_centro: Punto = top_center
"""Il punto centrale del lato superiore del rettangolo di delimitazione della grafica

:meta hide-value:
"""

basso_centro: Punto = bottom_center
"""Il punto centrale del lato inferiore del rettangolo di delimitazione della grafica

:meta hide-value:
"""

centro_sinistra: Punto = center_left
"""Il punto centrale del lato sinistro del rettangolo di delimitazione della grafica

:meta hide-value:
"""


centro_destra: Punto = center_right
"""Il punto centrale del lato destro del rettangolo di delimitazione della grafica

:meta hide-value:
"""


alto_sinistra: Punto = top_left
"""Il vertice in alto a sinistra del rettangolo di delimitazione della grafica

:meta hide-value:
"""

alto_destra: Punto = top_right
"""Il vertice in alto a destra del rettangolo di delimitazione della grafica

:meta hide-value:
"""

basso_sinistra: Punto = bottom_left
"""Il vertice in basso a sinistra del rettangolo di delimitazione della grafica

:meta hide-value:
"""

basso_destra: Punto = bottom_right
"""Il vertice in basso a destra del rettangolo di delimitazione della grafica

:meta hide-value:
"""



# pylint: disable=missing-docstring
# Separated from color.py only in the main English version to avoid circular imports

from pytamaro.color import Color
from pytamaro.checks import check_range


def rgb_color(red: int, green: int, blue: int, opacity: float = 1.0) -> Color:
    """
    Returns a color with the provided components for red (R), green (G) and blue (B) and a
    certain degree of opacity (alpha, A).

    .. figure:: https://upload.wikimedia.org/wikipedia/commons/thumb/b/b4/RGBCube_a.svg/524px-RGBCube_a.svg.png
       :height: 120px
       :align: center

       `RGB cube (SharkD via Wikimedia Commons) <https://commons.wikimedia.org/wiki/File:RGBCube_a.svg>`_

    :param red: red component [0-255]
    :param green: green component [0-255]
    :param blue: blue component [0-255]
    :param opacity: opacity (alpha) of the color, where 0 means fully
           transparent and 1 fully opaque. By default, all colors are fully opaque.
    :returns: a color with the provided RGBA components
    """
    check_range(red, 0, 255, "red")
    check_range(green, 0, 255, "green")
    check_range(blue, 0, 255, "blue")
    check_range(opacity, 0, 1, "opacity")
    return Color(red, green, blue, opacity)


def hsv_color(hue: float, saturation: float, value: float, opacity: float = 1.0) -> Color:
    """
    Returns a color with the provided hue (H), saturation (S), value (V) and a
    certain degree of opacity (alpha, A).

    .. figure:: https://upload.wikimedia.org/wikipedia/commons/4/4e/HSV_color_solid_cylinder.png
       :height: 120px
       :align: center

       `HSV cylinder (SharkD via Wikimedia Commons) <https://commons.wikimedia.org/wiki/File:HSV_color_solid_cylinder.png>`_

    :param hue: hue of the color [0-360]
    :param saturation: saturation of the color [0-1]
    :param value: the amount of light that is applied [0-1]
    :param opacity: opacity (alpha) of the color, where 0 means fully
           transparent and 1 fully opaque. By default, all colors are fully opaque.
    :returns: a color with the provided HSVA components.
    """
    check_range(hue, 0, 360, "hue")
    check_range(saturation, 0, 1, "saturation")
    check_range(value, 0, 1, "value")
    check_range(opacity, 0, 1, "opacity")
    chroma = value * saturation
    side = (hue / 60) % 6
    x_value = chroma * (1 - abs(side % 2 - 1))
    bottom_color = (chroma, x_value, 0)
    if 2 > side >= 1:
        bottom_color = (x_value, chroma, 0)
    if 3 > side >= 2:
        bottom_color = (0, chroma, x_value)
    if 4 > side >= 3:
        bottom_color = (0, x_value, chroma)
    if 5 > side >= 4:
        bottom_color = (x_value, 0, chroma)
    if side >= 5:
        bottom_color = (chroma, 0, x_value)
    to_add = value - chroma
    color = tuple(int((x + to_add) * 255) for x in bottom_color)
    return rgb_color(color[0], color[1], color[2], opacity)


def hsl_color(hue: float, saturation: float, lightness: float, opacity: float = 1.0) -> Color:
    """
    Returns a color with the provided hue (H), saturation (S), lightness (L) and a
    certain degree of opacity (alpha, A).

    .. figure:: https://upload.wikimedia.org/wikipedia/commons/3/35/HSL_color_solid_cylinder.png
       :height: 120px
       :align: center

       `HSL cylinder: SharkD via Wikimedia Commons <https://commons.wikimedia.org/wiki/File:HSL_color_solid_cylinder.png>`_

    :param hue: hue of the color [0-360]
    :param saturation: saturation of the color [0-1]
    :param lightness: the amount of white or black applied [0-1].
            Fully saturated colors have a lightness value of 1/2.
    :param opacity: opacity (alpha) of the color, where 0 means fully
           transparent and 1 fully opaque. By default, all colors are fully opaque.
    :returns: a color with the provided HSLA components
    """
    check_range(hue, 0, 360, "hue")
    check_range(saturation, 0, 1, "saturation")
    check_range(lightness, 0, 1, "lightness")
    check_range(opacity, 0, 1, "opacity")
    chroma = (1 - abs(2 * lightness - 1)) * saturation
    side = (hue / 60) % 6
    x_value = chroma * (1 - abs(side % 2 - 1))
    bottom_color = (chroma, x_value, 0)
    if 2 > side >= 1:
        bottom_color = (x_value, chroma, 0)
    if 3 > side >= 2:
        bottom_color = (0, chroma, x_value)
    if 4 > side >= 3:
        bottom_color = (0, x_value, chroma)
    if 5 > side >= 4:
        bottom_color = (x_value, 0, chroma)
    if side >= 5:
        bottom_color = (chroma, 0, x_value)
    to_add = lightness - chroma / 2
    color = tuple(int((x + to_add) * 255) for x in bottom_color)
    return rgb_color(color[0], color[1], color[2], opacity)



"""
Functions to enrich a graphic with information useful for debugging purposes.
"""

from skia import Point

from pytamaro.color import Color
from pytamaro.color_functions import rgb_color
from pytamaro.graphic import Graphic
from pytamaro.operations import (
    compose,
    graphic_height,
    graphic_width,
    overlay,
    pin,
    rotate,
)
from pytamaro.primitives import rectangle
from pytamaro.point_names import top_left, bottom_left


def top_left_point(graphic: Graphic) -> Point:
    """
    Returns the top left corner of the bounding box of the graphic.

    :param graphic: graphic to compute the top left corner of
    :returns: the top left corner of the bounding box
    """
    return graphic.bounds().toQuad()[0]


def add_debug_info(graphic: Graphic) -> Graphic:
    """
    Overlays debugging information onto a graphic (a border around it and a
    cross at the pin location).

    :param graphic: original graphic
    :returns: a graphic with debugging information
    :meta private:
    """
    relative_pin_pos = graphic.pin_position - top_left_point(graphic)
    border_thickness = 5
    border_color = rgb_color(240, 16, 16)
    g_with_border = add_border(graphic, border_thickness, border_color)
    new_rel_pin_pos = relative_pin_pos + (border_thickness, border_thickness)
    new_abs_pin_pos = top_left_point(g_with_border) + new_rel_pin_pos
    g_with_border.set_pin_position(new_abs_pin_pos.x(), new_abs_pin_pos.y())
    return show_pin_position(g_with_border)


def add_border(graphic: Graphic, width: float, color: Color) -> Graphic:
    """
    Adds a border with the given width (thickness) around the
    (bounding box of the) graphic.

    :param graphic: original graphic
    :param width: width of the border
    :param color: color of the border
    :returns: a new graphic with the border
    """
    horizontal = rectangle(graphic_width(graphic) + 2 * width, width, color)
    vertical = rectangle(width, graphic_height(graphic) + 2 * width, color)
    top_left_g = compose(pin(top_left, horizontal), pin(top_left, vertical))
    border = compose(
        pin(bottom_left, top_left_g), pin(bottom_left, rotate(180, top_left_g))
    )
    return overlay(graphic, border)


def show_pin_position(graphic: Graphic) -> Graphic:
    """
    Overlays a yellowish cross onto the graphic at its pin location,
    to make it visibile for debugging purposes.

    :param graphic: original graphic
    :returns: the graphic with the highlighted pin position
    """
    arm = rectangle(35, 5, rgb_color(250, 200, 0))
    cross = rotate(45, compose(arm, rotate(90, arm)))
    return compose(cross, graphic)



"""
Funktionen zum Erzeugen primitiver Grafiken (Figuren und Texte).
Falls nicht anders angegeben befindet sich die Fixierposition
in der Mitte des Begrenzungsrahmens der erzeugten Grafik.
"""

from __future__ import annotations

from pytamaro.de.color import Farbe
from pytamaro.de.graphic import Grafik
from pytamaro.primitives import (circular_sector, empty_graphic,
                                 rectangle, triangle)
from pytamaro.primitives import (ellipse as ellipse_e, text as text_e)


def rechteck(breite: float, hoehe: float, farbe: Farbe) -> Grafik:
    """
    Erzeugt ein Rechteck mit der gegebenen Breite und Höhe,
    gefüllt in der gegebenen Farbe.

    :param breite: die Breite des Rechtecks
    :param hoehe: die Höhe des Rechtecks
    :param farbe: Füllfarbe des Rechtecks
    :returns: eine Grafik mit dem gegebenen Rechteck
    """
    return rectangle(breite, hoehe, farbe)


def leere_grafik() -> Grafik:
    """
    Erzeugt eine leere Grafik.
    Wenn eine leere Grafik mit einer anderen Grafik kombiniert wird
    verhält sie sich als neutrales Element:
    das Ergebnis der Komposition ist einfach gleich der anderen Grafik.

    :returns: eine leere Grafik (Breite und Höhe sind 0)
    """
    return empty_graphic()


def ellipse(breite: float, hoehe: float, farbe: Farbe) -> Grafik:
    """
    Erzeugt eine Ellipse mit der gegebenen Breite und Höhe,
    gefüllt in der gegebenen Farbe.

    Wenn Breite und Höhe gleich gross sind wird die Ellipse zum Kreis
    mit dem entsprechenden Durchmesser.

    :param breite: Breite der Ellipse
    :param hoehe: Höhe der Ellipse
    :param farbe: Füllfarbe der Ellipse
    :returns: eine Grafik mit der gegebenen Ellipse
    """
    return ellipse_e(breite, hoehe, farbe)


def kreis_sektor(radius: float, winkel: float, farbe: Farbe) -> Grafik:
    """
    Erzeugt einen Kreissektor mit dem gegebenen Radius,
    der den gegebenen Winkel umspannt,
    gefüllt in der gegebenen Farbe.

    Ein Kreissektor ist ein Teil eines Kreises
    begrenzt durch zwei Radien und einen Bogen.
    Wenn man den Kreis als Uhr betrachtet
    dann zeigt der erste Radius in Richtung 3 Uhr.
    Der Winkel bestimmt die Position des zweiten Radius,
    ausgehend vom ersten Radius im Gegenuhrzeigersinn.
    Ein Winkel von 360 Grad entspricht einem ganzen Kreis.

    Die Fixierposition liegt in der Mitte des Kreises,
    aus dem der Kreissektor ausgeschnitten wurde.

    :param radius: Kreisradius
    :param winkel: Winkel des Sektors, in Grad
    :param farbe: Füllfarbe des Kreissektors
    :returns: eine Grafik mit dem gegebenen Kreissektor
    """
    return circular_sector(radius, winkel, farbe)


def dreieck(seite1: float, seite2: float, winkel: float, farbe: Farbe) -> Grafik:
    """
    Erzeugt ein Dreieck mit den gegebenen Seitenlängen und dem gegebenen Winkel,
    gefüllt in der gegebenen Farbe.
    Die erste Seite verläuft horizontal nach rechts.
    Der Winkel gibt an, wie viel die zweite Seite
    im Gegenuhrzeigersinn von der ersten Seite abweicht.

    Für alle Dreiecke, ausser für stumpfe Dreiecke,
    liegt die untere linke Ecke des Begrenzungsrahmens
    auf dem Eckpunkt des Dreiecks, für das der Winkel angegeben ist.

    Die Fixierposition liegt auf dem Schwerpunkt des Dreiecks.

    :param seite1: Länge der ersten, horizontalen Seite
    :param seite2: Länge der zweiten Seite
    :param winkel: Winkel von der ersten zu der zweiten Seiten, in Grad
    :param farbe: Farbe des Dreiecks
    :returns: eine Grafik mit dem gegebenen Dreieck
    """
    return triangle(seite1, seite2, winkel, farbe)


def text(inhalt: str, schriftart: str, punkte: float, farbe: Farbe) -> Grafik:
    """
    Erzeugt einen Text in der gegebenen Schriftart und Schriftgrösse,
    gefüllt in der gegebenen Farbe.

    Falls für die gegebene Schriftart auf dem System keine True-Type Schrift
    zur Verfügung steht, wird eine einfache Standardschriftart verwendet.
    Die resultierende Grafik hat die minimale Grösse,
    die den gesamten Text umschliesst.

    Die Fixierposition liegt auf der linken Kante des Begrenzungsrahmens,
    auf der Höhe der Grundlinie des Textes.

    :param inhalt: der Text, der dargestellt werden soll
    :param schriftart: der Name der Schriftart
                       (zum Beispiel "arial" auf Windows, "Arial" auf macOS)
    :param punkte: Schriftgrösse in typografischen Punkten (zum Beispiel 16)
    :param farbe: Farbe, in der der Text dargestellt werden soll
    :returns: eine Grafik bestehend aus dem gegebenen Text
    """
    return text_e(inhalt, schriftart, punkte, farbe)



"""
Functions to create primitive graphics (shapes and text).
Unless specified otherwise, the initial pinning position is at the center of the
graphic's bounding box.
"""

from pytamaro.color import Color
from pytamaro.graphic import (CircularSector, Ellipse, Empty, Graphic,
                              Rectangle, Text, Triangle)
from pytamaro.checks import check_angle, check_color, check_length, check_number, check_type
from pytamaro.utils import export


@export
def rectangle(width: float, height: float, color: Color) -> Graphic:
    """
    Creates a rectangle of the given size, filled with a color.

    :param width: width of the rectangle
    :param height: height of the rectangle
    :param color: the color to be used to fill the rectangle
    :returns: the specified rectangle as a graphic
    """
    check_length(width, "width")
    check_length(height, "height")
    check_color(color)
    return Rectangle(width, height, color)


@export
def empty_graphic() -> Graphic:
    """
    Creates an empty graphic.
    When an empty graphic is composed with any other graphic, it behaves
    as a neutral element: the result is always identical to the other graphic.

    :returns: an empty graphic (width and height 0)
    """
    return Empty()


@export
def ellipse(width: float, height: float, color: Color) -> Graphic:
    """
    Creates an ellipse with the given width and height, filled with a color.

    When width and height are the same, the ellipse becomes a circle with a
    diameter equal to the provided size.

    :param width: width of the ellipse
    :param height: height of the ellipse
    :param color: the color to be used to fill the circle
    :returns: the specified circle as a graphic
    """
    check_length(width, "width")
    check_length(height, "height")
    check_color(color)
    return Ellipse(width, height, color)


@export
def circular_sector(radius: float, angle: float, color: Color) \
        -> Graphic:
    """
    Creates a circular sector belonging to a circle of the given radius, filled
    with a color.

    A circular sector is a portion of a circle enclosed between two radii and
    an arc.
    Considering a circle as a clock, the first radius is supposed to "point"
    towards 3 o'clock. The `angle` determines the position of the second
    radius, computed starting from the first one in counterclockwise direction.
    An angle of 360 degrees corresponds to a full circle.

    The pinning position is at the center of the circle from which the circular
    sector is taken.

    :param radius: radius of the circle from which the circular sector is
                   taken
    :param angle: central angle, in degrees
    :param color: the color to be used to fill the circular sector
    :returns: the specified circular sector as a graphic
    """
    check_length(radius, "radius")
    check_angle(angle, 0, 360)
    check_color(color)
    return CircularSector(radius, angle, color)


@export
def triangle(side1: float, side2: float, angle: float, color: Color) -> Graphic:
    """
    Creates a triangle specifying two sides and the angle between them, filled
    with a color.
    The first side extends horizontally to the right. The angle specifies how
    much the second side is rotated, counterclockwise, from the first one.

    For all triangles, except obtuse ones, the bottom-left corner of the
    resulting graphic concides with the vertex of the triangle for which the
    angle is specified.

    The pinning position is the centroid of the triangle.

    :param side1: length of the first, horizontal side of the triangle
    :param side2: length of the second side of the triangle
    :param angle: angle between the two sides, in degrees
    :param color: the color to be used to fill the triangle
    :returns: the specified triangle as a graphic
    """
    check_length(side1, "side1")
    check_length(side2, "side2")
    check_angle(angle, 0, 180)
    return Triangle(side1, side2, angle, color)


@export
def text(content: str, font: str, points: float, color: Color) \
        -> Graphic:
    """
    Creates a graphic with the text rendered using the specified font, size and
    color.

    When the indicated True-Type Font is not found in the system, a very
    basilar font that is always available is used instead. The resulting
    graphic has the minimal size that still fits the whole text.

    The pinning position is horizontally aligned on the left and vertically on
    the baseline of the text.

    :param content: the text to render
    :param font: the name of the font (e.g., "arial" on Windows, "Arial" on
           macOS)
    :param points: size in typographic points (e.g., 16)
    :param color: the color to be used to render the text
    :returns: the specified text as a graphic
    """
    check_type(content, str, "content")
    check_type(font, str, "font")
    check_number(points, "points")
    check_color(color)
    return Text(content, font, points, color)



"""
Functions to do operations on graphics (mainly, to combine them).
"""


from pytamaro.graphic import Compose, Graphic, Pin, Rotate
from pytamaro.checks import check_angle, check_graphic, check_point
from pytamaro.utils import export
from pytamaro.point import Point
from pytamaro.point_names import center, top_center, bottom_center, center_right, center_left


@export
def graphic_width(graphic: Graphic) -> int:
    """
    Returns the width of a graphic.

    :param graphic: graphic to calculate the width of
    :returns: width of the graphic
    """
    check_graphic(graphic)
    return graphic.size().toCeil().width()


@export
def graphic_height(graphic: Graphic) -> int:
    """
    Returns the height of a graphic.

    :param graphic: graphic to calculate the height of
    :returns: height of the graphic
    """
    check_graphic(graphic)
    return graphic.size().toCeil().height()


@export
def compose(foreground_graphic: Graphic, background_graphic: Graphic) -> Graphic:
    """
    Creates a new graphic by composing the two provided graphics.
    The first graphic is kept in the foreground, the second one in the
    background.
    The graphics are aligned by superimposing their pinning positions.

    The pinning position used to compose becomes the pinning position of the
    resulting graphic.

    :param foreground_graphic: graphic in the foreground
    :param background_graphic: graphic in the background
    :returns: the resulting composed graphic
    """
    check_graphic(foreground_graphic, "foreground_graphic")
    check_graphic(background_graphic, "background_graphic")
    return Compose(foreground_graphic, background_graphic)


@export
def pin(point: Point, graphic: Graphic) -> Graphic:
    """
    Creates a new graphic that corresponds to the provided graphic,
    with a new pinning position.

    Each graphic is contained in a rectangular bounding box.
    There are 9 notable points, corresponding to the four corners of this rectangle,
    the middle points of the four edges and the center of the rectangle.
    These points can be referred to using these names: `top_left`, `top_right`,
    `bottom_left`, `bottom_right`, `top_center`, `center_right`,
    `bottom_center`, `center_left` and `center`.

    :param point: the point indicating the new pinning position
    :param graphic: original graphic
    :returns: a new graphic with the specified pinning position
    """
    check_point(point)
    check_graphic(graphic)
    return Pin(graphic, point)


def _compose_pin_center(graphic1: Graphic, graphic2: Graphic,
                        point1: Point, point2: Point) -> Graphic:
    """
    Composes the two graphics, pinning the first one at `point1` and the second one
    at `point2`. The composed graphic is then pinned at its center.

    :param graphic1: first graphic to compose
    :param graphic2: second graphic to compose
    :param point1: pinning position for the first graphic
    :param point2: pinning position for the second graphic

    :returns: the combined graphic, pinned at its center
    """
    return pin(center, compose(pin(point1, graphic1), pin(point2, graphic2)))


@export
def overlay(foreground_graphic: Graphic, background_graphic: Graphic) -> Graphic:
    """
    Creates a new graphic by overlaying the two provided graphics,
    keeping the first one in the foreground and the second one in background.
    The two graphics are overlaid on their centers.

    The pinning position of the new graphic is at its center.

    :param foreground_graphic: graphic in the foreground
    :param background_graphic: graphic in the background
    :returns: the resulting graphic after overlaying the two provided ones
    """
    check_graphic(foreground_graphic, "foreground_graphic")
    check_graphic(background_graphic, "background_graphic")
    return _compose_pin_center(foreground_graphic, background_graphic, center, center)


@export
def beside(left_graphic: Graphic, right_graphic: Graphic) -> Graphic:
    """
    Creates a new graphic by placing the two graphics one besides the other.
    The two graphics are vertically centered.

    The pinning position of the new graphic is at its center.

    :param left_graphic: graphic to place on the left
    :param right_graphic: graphic to place on the right
    :returns: the resulting graphic after placing the two graphics one besides
              the other
    """
    check_graphic(left_graphic, "left_graphic")
    check_graphic(right_graphic, "right_graphic")
    return _compose_pin_center(left_graphic, right_graphic, center_right, center_left)


@export
def above(top_graphic: Graphic, bottom_graphic: Graphic) -> Graphic:
    """
    Creates a new graphic by placing the two graphics one above the other.
    The two graphics are horizontally centered.

    The pinning position of the new graphic is at its center.

    :param top_graphic: graphic to place on the top
    :param bottom_graphic: graphic to place on the bottom
    :returns: the resulting graphic after placing the two graphics one above
              the other
    """
    check_graphic(top_graphic, "top_graphic")
    check_graphic(bottom_graphic, "bottom_graphic")
    return _compose_pin_center(top_graphic, bottom_graphic, bottom_center, top_center)


@export
def rotate(angle: float, graphic: Graphic) -> Graphic:
    """
    Creates a new graphic by rotating counterclockwise the provided graphic
    around its pinning position by the given angle.
    A negative angle corresponds to a clockwise rotation.

    :param angle: angle of counterclockwise rotation, in degrees
    :param graphic: the graphic to rotate
    :returns: a new, rotated graphic
    """
    check_angle(angle)
    check_graphic(graphic)
    # Negate the angle given that Rotate is clockwise.
    return Rotate(graphic, -angle)



"""
`Color` type, functions to produce colors, and constants for important colors.
"""

from dataclasses import dataclass
from typing import Tuple

from skia import Color4f


@dataclass
class Color:
    """
    Represents a color.
    A color also has a degree of opacity,
    from completely transparent (like the color `transparent`)
    to completely opaque (like the color `red`).
    """

    color: Color4f

    def __init__(self, red: int, green: int, blue: int, alpha: float):
        self.color = Color4f(red / 255, green / 255, blue / 255, alpha)

    def as_tuple(self) -> Tuple[int, int, int, float]:
        """
        Returns the current color as an RGBA tuple.

        :meta private:
        :returns: a tuple with four components. The first three [0-255] identify the
                  color, the last one [0-1] identifies the transparency
        """
        return self.color[0] * 255, self.color[1] * 255, self.color[2] * 255, self.color[3]



"""
Funktionen für Operationen mit Grafiken (hauptsächlich für deren Komposition).
"""

from __future__ import annotations

from pytamaro.de.graphic import Grafik
from pytamaro.de.point import Punkt
from pytamaro.operations import (above, beside, compose, graphic_height,
                                 graphic_width, overlay, pin, rotate)


def grafik_breite(grafik: Grafik) -> int:
    """
    Gibt die Breite der gegebenen Grafik zurück.

    :param grafik: Grafik deren Breite gesucht ist
    :returns: Breite der Grafik
    """
    return graphic_width(grafik)


def grafik_hoehe(grafik: Grafik) -> int:
    """
    Gibt die Höhe der gegebenen Grafik zurück.

    :param grafik: Grafik deren Höhe gesucht ist
    :returns: Höhe der Grafik
    """
    return graphic_height(grafik)


def kombiniere(vordere_grafik: Grafik, hintere_grafik: Grafik) \
        -> Grafik:
    """
    Erzeugt eine neue Grafik,
    die aus der Kombination der zwei gegebenen Grafiken besteht.
    Die erste gegebene Grafik liegt im Vordergrund
    und die zweite im Hintergrund.
    Die Grafiken werden so ausgerichtet,
    dass ihre Fixierpositionen übereinanderliegen.

    Die überlappenden Fixierpositionen werden zur Fixierposition der
    resultierenden Grafik.

    :param vordere_grafik: Grafik im Vordergrund
    :param hintere_grafik: Grafik im Hintergrund
    :returns: die zusammengesetzte Grafik
    """
    return compose(vordere_grafik, hintere_grafik)


def fixiere(punkt: Punkt, grafik: Grafik) -> Grafik:
    """
    Erzeugt eine neue Grafik,
    die der gegebenen Grafik mit einer neuen Fixierposition entspricht.

    Jede Grafik liegt in einem rechteckigen Begrenzungsrahmen.
    Der Rahmen definiert 9 nennenswerte Punkte:
    die vier Ecken, die Mittelpunkte der vier Kanten und die Mitte des Rahmens.
    Die Namen dieser Punkte sind: `oben_links`, `oben_mitte`, `oben_rechts`,
    `mitte_links`, `mitte`, `mitte_rechts`, `unten_links`, `unten_mitte` und `unten_rechts`.

    :param point: ein Punkt welcher die neue Fixierposition bestimmt
    :param graphic: die ursprüngliche Grafik
    :returns: eine neue Grafik mit der gegebenen Fixierposition
    """
    return pin(punkt, grafik)


def ueberlagere(vordere_grafik: Grafik, hintere_grafik: Grafik) \
        -> Grafik:
    """
    Erzeugt eine neue Grafik,
    die aus der zentrierten Überlagerung der zwei gegebenen Grafiken besteht.
    Die erste gegebene Grafik liegt im Vordergrund
    und die zweite im Hintergrund.

    Die Fixierposition der neuen Grafik liegt in deren Mitte.

    :param vordere_grafik: Grafik im Vordergrund
    :param hintere_grafik: Grafik im Hintergrund
    :returns: die zusammengesetzte Grafik
    """
    return overlay(vordere_grafik, hintere_grafik)


def neben(linke_grafik: Grafik, rechte_grafik: Grafik) -> Grafik:
    """
    Erzeugt eine neue Grafik,
    die aus dem Nebeneinanderlegen der zwei gegebenen Grafiken besteht.
    Die zwei Grafiken sind vertikal zentriert.

    Die Fixierposition der neuen Grafik liegt in deren Mitte.

    :param linke_grafik: linke Grafik (im Westen)
    :param rechte_grafik: rechte Grafik (im Osten)
    :returns: die zusammengesetzte Grafik
    """
    return beside(linke_grafik, rechte_grafik)


def ueber(obere_grafik: Grafik, untere_grafik: Grafik) -> Grafik:
    """
    Erzeugt eine neue Grafik,
    die aus dem Übereinanderlegen der zwei gegebenen Grafiken besteht.
    Die zwei Grafiken sind horizontal zentriert.

    Die Fixierposition der neuen Grafik liegt in deren Mitte.

    :param obere_grafik: obere Grafik (im Norden)
    :param untere_grafik: untere Grafik (im Süden)
    :returns: die zusammengesetzte Grafik
    """
    return above(obere_grafik, untere_grafik)


def drehe(winkel: float, grafik: Grafik) -> Grafik:
    """
    Erzeugt eine neue Grafik, die einer Rotation der gegebenen Grafik
    um ihre Fixierposition im Gegenuhrzeigersinn
    um den gegebenen Winkel entspricht.
    Negative Winkel entsprechen einer Rotation um Uhrzeigersinn.

    :param winkel: Drehwinkel, in Grad im Gegenuhrzeigersinn
    :param grafik: zu rotierende Grafik
    :returns: die neue, rotierte Grafik
    """
    return rotate(winkel, grafik)



"""
Funzioni per output di grafiche (visualizzare o salvare).
"""

from __future__ import annotations

from pytamaro.io import save_animation, save_graphic, show_animation, show_graphic
from pytamaro.it.graphic import Grafica


def visualizza_grafica(grafica: Grafica, debug: bool = False):
    """
    Visualizza una grafica. Grafiche prive di area non possono essere
    visualizzate.

    Quando `debug` è `True`, adorna la visualizzazione con informazioni utili
    per debugging: un bordo rosso attorno alla bounding box e una croce
    giallastra attorno al punto di fissaggio.

    :param grafica: grafica da visualizzare
    :param debug: può facoltativamente essere impostato a `True` per
           sovrapporre informazioni di debug
    """
    show_graphic(grafica, debug)


def salva_grafica(nome_file: str, grafica: Grafica, debug: bool = False):
    """
    Salva una grafica in un file.
    Due formati di file sono supportati: PNG (grafica raster) e SVG (grafica
    vettoriale).
    L'estensione del nome del file (o ".png" o ".svg") determina il formato.

    Una grafica priva di area non può essere salvata nel formato PNG.

    Quando `debug` è `True`, adorna la visualizzazione con informazioni utili
    per debugging: un bordo rosso attorno alla bounding box e una croce
    giallastra attorno al punto di fissaggio.

    :param nome_file: nome del file da creare (con l'estensione)
    :param grafica: grafica da visualizzare
    :param debug: può facoltativamente essere impostato a `True` per
           sovrapporre informazioni di debug
    """
    save_graphic(nome_file, grafica, debug)


def salva_animazione(
    nome_file: str, grafiche: list[Grafica], durata: int = 40, loop: bool = True
):
    """
    Salva una sequenza di grafiche come un'animazione (GIF).

    Le grafiche vengono riprodotte sequenzialmente (normalmente a 25 frame al
    secondo) a ciclo continuo.

    :param nome_file: nome del file da creare (inclusa l'estensione '.gif')
    :param grafiche: lista di grafiche da salvare come animazione
    :param durata: durata in millisecondi di ciascun frame (default a 40
           millisecondi, ovvero 25 frame al secondo)
    :param loop: determina se la GIF debba riprodursi in loop indefinitamente (default a True)
    """
    save_animation(nome_file, grafiche, durata, loop)


def visualizza_animazione(grafiche: list[Grafica], durata: int = 40, loop: bool = True):
    """
    Visualizza una sequenza di grafiche come un'animazione (GIF).

    Le grafiche vengono riprodotte sequenzialmente (normalmente a 25 frame al
    secondo) a ciclo continuo.

    :param grafiche: lista di grafiche da salvare come animazione
    :param durata: durata in millisecondi di ciascun frame (default a 40
           millisecondi, ovvero 25 frame al secondo)
    :param loop: determina se la GIF debba riprodursi in loop indefinitamente (default a True)
    """
    show_animation(grafiche, durata, loop)



"""
`Point` type.
"""
from __future__ import annotations

from dataclasses import dataclass


@dataclass
class Point:
    """
    Represents a point on a plane.
    """
    x: float    # pylint: disable=invalid-name
    y: float    # pylint: disable=invalid-name

    def translate(self, current_vector: Vector) -> Point:
        """
        Returns a point obtained by translating `current_point` using `current_vector`

        :meta private:
        :param current_point: the point to be translated
        :param current_vector: the vector used for translation
        :returns: a new point, the result of the translation
        """
        return Point(self.x + current_vector.terminal_point.x,
                     self.y + current_vector.terminal_point.y)


# Center of the 2-dimensional space
zero = Point(0.0, 0.0)


@dataclass
class Vector:
    """
    Represents a vector in 2-dimensional space storing its terminal point,
    assuming it starts at the origin (0, 0).

    :meta private:
    """
    terminal_point: Point

    def __add__(self, other: Vector):
        """
        Returns the vector obtained by adding `other` to the current one

        :param other: the vector added to the current one
        :returns: a vector that represent the sum of the current one and `other`.
        """
        return Vector(Point(self.terminal_point.x + other.terminal_point.x,
                            self.terminal_point.y + other.terminal_point.y))

    def __mul__(self, factor: float):
        """
        Returns this vector multiplied by a given scalar

        :param factor: scalar used in the multiplication
        :returns: a new vector, the result of scalar-vector multiplication
        """
        return Vector(Point(self.terminal_point.x * factor, self.terminal_point.y * factor))


# Canonical basis of R2
i_hat = Vector(Point(1.0, 0.0))
j_hat = Vector(Point(0.0, 1.0))



"""
Nomi di colori notevoli essendo ai vertici del cubo RGB.
In aggiunta, il colore completamente trasparente.
"""

from __future__ import annotations

from pytamaro.color_names import black, blue, cyan, green, magenta, red, transparent, white, yellow
from pytamaro.it.color import Colore

nero: Colore = black
"""Colore nero

:meta hide-value:
"""

rosso: Colore = red
"""Colore rosso

:meta hide-value:
"""

verde: Colore = green
"""Colore verde

:meta hide-value:
"""

blu: Colore = blue
"""Colore blu

:meta hide-value:
"""

giallo: Colore = yellow
"""Colore giallo

:meta hide-value:
"""

magenta: Colore = magenta  # pylint: disable=self-assigning-variable
"""Colore magenta

:meta hide-value:
"""

ciano: Colore = cyan
"""Colore ciano

:meta hide-value:
"""

bianco: Colore = white
"""Colore bianco

:meta hide-value:
"""

trasparente: Colore = transparent
"""Colore completamente trasparente

:meta hide-value:
"""



"""
(Re-)Export German names for:
- basic types (Graphic and Color)
- color names
- operations
- primitves
- I/O
from this top-level module.
"""


import sys

from pytamaro.de.color import *
from pytamaro.de.color_names import *
from pytamaro.de.graphic import *
from pytamaro.de.io import *
from pytamaro.de.operations import *
from pytamaro.de.primitives import *
from pytamaro.de.point_names import *

setattr(sys.modules["pytamaro"], "LANGUAGE", "de")



"""
Funzioni per creare grafiche primitive (forme e testo).
Tranne quando specificato diversamente, la posizione di fissaggio iniziale è al
centro del rettangolo di delimitazione (bounding box) della grafica.
"""

from __future__ import annotations

from pytamaro.it.color import Colore
from pytamaro.it.graphic import Grafica
from pytamaro.primitives import (circular_sector, ellipse, empty_graphic,
                                 rectangle, text, triangle)


def rettangolo(larghezza: float, altezza: float, colore: Colore) -> Grafica:
    """
    Crea un rettangolo delle dimensioni indicate, riempito con un colore.

    :param larghezza: larghezza del rettangolo
    :param altezza: altezza del rettangolo
    :param colore: colore da usare per riempire il rettangolo
    :returns: una grafica con il rettangolo specificato
    """
    return rectangle(larghezza, altezza, colore)


def grafica_vuota() -> Grafica:
    """
    Crea una grafica vuota.
    Quando una grafica vuota viene composta con ogni altra grafica, si comporta
    da elemento neutro: il risultato è sempre uguale all'altra grafica.

    :returns: una grafica vuota (larghezza e altezza 0)
    """
    return empty_graphic()


def ellisse(larghezza: float, altezza: float, colore: Colore) -> Grafica:
    """
    Crea un ellisse delle dimensioni indicate, riempito con un colore.

    Quando larghezza e altezza coincidono, l'ellisse diventa un cerchio di
    diametro pari alla dimensione indicata.

    :param larghezza: larghezza dell'ellisse
    :param altezza: altezza dell'ellisse
    :param colore: colore da usare per riempire l'ellisse
    :returns: una grafica con l'ellisse specificato
    """
    return ellipse(larghezza, altezza, colore)


def settore_circolare(raggio: float, angolo: float, colore: Colore) -> Grafica:
    """
    Crea un settore circolare appartenente a un cerchio del raggio indicato,
    riempito con un colore.

    Un settore circolare è una porzione di cerchio racchiusa tra due raggi e un
    arco.
    Considerando il cerchio come un orologio, il primo raggio "punta" in
    direzione delle ore 3. L'`angolo` determina la posizione del secondo
    raggio, calcolata a partire dalla posizione del primo in senso antiorario.
    Un angolo di 360 gradi corrisponde a un cerchio completo.

    La posizione di fissaggio è al centro del cerchio da cui è preso il settore
    circolare.

    :param raggio: raggio del cerchio da cui è preso il settore circolare
    :param angolo: angolo al centro, in gradi
    :param colore: colore da usare per riempire il settore circolare
    :returns: una grafica con il settore circolare specificato
    """
    return circular_sector(raggio, angolo, colore)


def triangolo(lato1: float, lato2: float, angolo: float, colore: Colore) -> Grafica:
    """
    Crea un triangolo specificando due lati e l'angolo tra essi compreso,
    riempito con un colore.
    Il primo lato si estende orizzontalmente verso destra. L'angolo specifica
    di quanto il secondo lato è ruotato, in senso antiorario, rispetto al primo.

    Per tutti i triangoli, eccetto quelli ottusi, il punto in basso a sinistra
    della grafica risultante coincide con il vertice del triangolo di cui si è
    specificato l'angolo.

    La posizione di fissaggio è il centroide del triangolo.

    :param lato1: lunghezza del primo lato (orizzontale) del triangolo
    :param lato1: lunghezza del secondo lato del triangolo
    :param angolo: angolo compreso tra i due lati, in gradi
    :param colore: colore da usare per riempire il triangolo
    :returns: una grafica con il triangolo specificato
    """
    return triangle(lato1, lato2, angolo, colore)


def testo(contenuto: str, font: str, punti: float, colore: Colore) -> Grafica:
    """
    Crea una grafica con il testo renderizzato usando font, dimensione e colore
    indicati.

    Quando il font True-Type indicato non è disponibile nel sistema, al suo
    posto viene usato un font estremamente basilare e sempre disponibile. La
    grafica risultante ha la dimensione minima in modo da racchiudere l'intero
    testo.

    La posizione di fissaggio è allineata orizzontalmente a sinistra e
    verticalmente sulla linea di base (baseline) del testo.

    :param contenuto: il testo di cui fare rendering
    :param font: il nome del font (ad esempio "arial" su Windows, "Arial" su
           macOS)
    :param punti: dimensione in punti tipografici (ad esempio 16)
    :param colore: colore da usare per fare il rendering del testo
    :returns: una grafica con il testo specificato
    """
    return text(contenuto, font, punti, colore)



"""
Tipo `Grafica`, che include una grafica con una posizione di fissaggio.
"""

from __future__ import annotations

from pytamaro.graphic import Graphic

Grafica = Graphic
"""
Una grafica (immagine) con una posizione per fissare.

La posizione di fissaggio viene usata nelle seguenti operazioni:

- rotazione (per determinare il centro di rotazione)
- composizione di grafiche (due grafiche vengono composte allineando le loro
  posizioni di fissaggio).
"""



"""
Names of notable points, that can be used as pinning positions for a graphic.
"""

from pytamaro.point import zero, i_hat, j_hat, Point


center: Point = zero
"""The center point of the bounding box

:meta hide-value:
"""


top_center: Point = center.translate(j_hat)
"""The middle point of the top edge of the bounding box

:meta hide-value:
"""

bottom_center: Point = center.translate(j_hat * -1)
"""The middle point of the bottom edge of the bounding box

:meta hide-value:
"""

center_left: Point = center.translate(i_hat * -1)
"""The middle point of the left edge of the bounding box

:meta hide-value:
"""


center_right: Point = center.translate(i_hat)
"""The middle point of the right edge of the bounding box

:meta hide-value:
"""


top_left: Point = center_left.translate(j_hat)
"""The top left corner of the bounding box

:meta hide-value:
"""

top_right: Point = center_right.translate(j_hat)
"""The top right corner of the bounding box

:meta hide-value:
"""


bottom_left: Point = center_left.translate(j_hat * -1)
"""The bottom left corner of the bounding box

:meta hide-value:
"""


bottom_right: Point = center_right.translate(j_hat * -1)
"""The bottom right corner of the bounding box

:meta hide-value:
"""



"""
Tipo `Colore`, funzioni per produrre colori e costanti per colori notevoli.
"""

from __future__ import annotations

from pytamaro.color import Color
from pytamaro.color_functions import rgb_color, hsl_color, hsv_color

Colore = Color
"""
Rappresenta un colore.
Un colore ha anche un grado di opacità,
da completamente trasparente (come il colore `trasparente`)
a completamente opaco (come il colore `rosso`).
"""


def colore_rgb(rosso: int, verde: int, blu: int, opacita: float = 1.0) -> Colore:
    """
    Ritorna un colore con le componenti indicate per rosso (R), verde (G) e blu (B) e un
    certo grado di opacità (alpha, A).

    .. figure:: https://upload.wikimedia.org/wikipedia/commons/thumb/b/b4/RGBCube_a.svg/524px-RGBCube_a.svg.png
       :height: 120px
       :align: center

       `Cubo RGB (SharkD via Wikimedia Commons) <https://commons.wikimedia.org/wiki/File:RGBCube_a.svg>`_

    :param rosso: componente rosso [0-255]
    :param verde: componente verde [0-255]
    :param blu: componente blu [0-255]
    :param opacita: opacità (alpha) del colore, dove 0 significa completamente
           trasparente e 1 completamente opaco.
           Di default, tutti i colori sono completamente opachi.
    :returns: un colore con le componenti RGBA indicate
    """
    return rgb_color(rosso, verde, blu, opacita)


def colore_hsv(tonalita: float, saturazione: float, valore: float, opacita: float = 1.0) -> Colore:
    """
    Ritorna un colore con la tonalità (H), saturazione (S) e valore (V) indicati,
    e un certo grado di opacità (alpha, A).

    .. figure:: https://upload.wikimedia.org/wikipedia/commons/4/4e/HSV_color_solid_cylinder.png
       :height: 120px
       :align: center

       `Cilindro HSV (SharkD via Wikimedia Commons) <https://commons.wikimedia.org/wiki/File:HSV_color_solid_cylinder.png>`_

    :param tonalita: tonalità del colore [0-360]
    :param saturazione: saturazione del colore [0-1]
    :param valore: quantità di luce applicata [0-1]
    :param opacita: opacità (alpha) del colore, dove 0 significa completamente
           trasparente e 1 completamente opaco.
           Di default, tutti i colori sono completamente opachi.
    :returns: un colore con le componenti HSVA indicate
    """
    return hsv_color(tonalita, saturazione, valore, opacita)


def colore_hsl(tonalita: float, saturazione: float, luce: float, opacita: float = 1.0) -> Colore:
    """
    Ritorna un colore con la tonalità (H), saturazione (S) e luce (L) indicati,
    e un certo grado di opacità (alpha, A).

    .. figure:: https://upload.wikimedia.org/wikipedia/commons/3/35/HSL_color_solid_cylinder.png
       :height: 120px
       :align: center

       `Cilindro HSL: SharkD via Wikimedia Commons <https://commons.wikimedia.org/wiki/File:HSL_color_solid_cylinder.png>`_

    :param tonalita: tonalità del colore [0-360]
    :param saturazione: saturazione del colore [0-1]
    :param luce: quantità di bianco o nero applicata [0-1].
           Colori completamente saturi hanno un valore di luce di 1/2.
    :param opacita: opacità (alpha) del colore, dove 0 significa completamente
           trasparente e 1 completamente opaco.
           Di default, tutti i colori sono completamente opachi.
    :returns: un colore con le componenti HSLA indicate
    """
    return hsl_color(tonalita, saturazione, luce, opacita)



"""
Checks to be performed on the parameters of user-facing functions.
"""

from math import inf
from numbers import Number
from typing import Any

from pytamaro.color import Color
from pytamaro.graphic import Graphic

from pytamaro.localization import translate
from pytamaro.point import Point


def check_angle(angle: Any, lower_bound: float = -inf, upper_bound: float = inf):
    """
    Raises an exception when the provided value is not valid for a
    angle, being outside the specified range or not a numeric type.

    :param angle: the value for an angle to be checked
    """
    check_range(angle, lower_bound, upper_bound, "angle")


def check_length(length: Any, parameter_name: str):
    """
    Raises an exception when the provided value is not valid for a
    length, being negative or not a numeric type.

    :param length: the value for a length to be checked
    """
    check_number(length, parameter_name)
    localized_parameter_name = translate(parameter_name)
    if length < 0:
        raise ValueError(translate("INVALID_LENGTH", localized_parameter_name))


def check_type(value: Any, expected_type: type, parameter_name: str):
    """
    Raises an exception when the provided value is not valid for a
    given type.

    :param value: the value to be checked
    :param expected_type: the expected type for the value
    :param parameter_name: original parameter name, to be used in the error message
    """
    if not isinstance(value, expected_type):
        expected_type_name: str = expected_type.__name__
        actual_type_name: str = type(value).__name__
        raise TypeError(translate("INVALID_TYPE",
                                  translate(parameter_name),
                                  translate(expected_type_name),
                                  translate(actual_type_name)))


def check_color(color: Any):
    """
    Raises an exception when the provided value is not valid for a
    color, not being of type Color.

    :param color: the value for a color to be checked
    """
    check_type(color, Color, "color")


def check_graphic(graphic: Any, parameter_name: str = "graphic"):
    """
    Raises an exception when the provided value is not valid for a
    graphic, not being of type Graphic.

    :param color: the value for a color to be checked
    """
    check_type(graphic, Graphic, parameter_name)


def check_point(point: Any):
    """
    Raises an exception when the provided value is not valid for a
    point, not being of type Point.

    :param point: the value for a point to be checked
    """
    check_type(point, Point, "point")


def check_number(value: Any, parameter_name: str):
    """
    Raises an exception when the provided value is not valid for a
    number, not being of type Number.

    :param value: the value to be checked
    :param parameter_name: original parameter name, to be used in the error message
    """
    check_type(value, Number, parameter_name)


def check_range(value: Any, lower_bound: float, upper_bound: float, parameter_name: str):
    """
    Raises an exception when the provided value is not valid for a
    range, being outside the specified range or not a numeric type.

    :param value: the value to be checked
    :param lower_bound: the lower bound of the range
    :param upper_bound: the upper bound of the range
    """
    check_number(value, parameter_name)
    if value < lower_bound or value > upper_bound:
        raise ValueError(translate("INVALID_RANGE",
                                   translate(parameter_name),
                                   lower_bound,
                                   upper_bound))



"""
Funktionen zur Ausgabe (Anzeigen oder Speichern) von Grafiken.
"""

from __future__ import annotations

from pytamaro.io import save_animation, save_graphic, show_animation, show_graphic
from pytamaro.de.graphic import Grafik


def zeige_grafik(grafik: Grafik, debug: bool = False):
    """
    Zeige die gegebene Grafik an.

    Eine Grafik ohne Fläche kann nicht angezeigt werden.

    Falls `debug` `True` ist werden auf der Grafik zusätzliche Informationen
    dargestellt, welche für das Debugging nützlich sein können.
    Ein roter Rahmen markiert den Begrenzungsrahmen der Grafik,
    und ein gelbliches Kreuz gibt die Fixierposition an.

    :param grafik: die anzuzeigende Grafik
    :param debug: kann optional auf `True` gesetzt werden, um über der Grafik
                  Debug-Informationen darzustellen
    """
    show_graphic(grafik, debug)


def speichere_grafik(datei_name: str, grafik: Grafik, debug: bool = False):
    """
    Speichere die gegebene Grafik in einer Datei.

    Zwei Dateiformate werden unterstützt: PNG (Rastergrafik) und SVG (Vektorgrafik).
    Die Dateinamenerweiterung (entweder ".png" oder ".svg") bestimmt das Dateiformat.

    Grafiken ohne Fläche können nicht als PNG-Datei gespeichert werden.

    Falls `debug` `True` ist werden auf der Grafik zusätzliche Informationen
    dargestellt, welche für das Debugging nützlich sein können.
    Ein roter Rahmen markiert den Begrenzungsrahmen der Grafik,
    und ein gelbliches Kreuz gibt die Fixierposition an.

    :param datei_name: Name der zu kreierenden Datei (mit Erweiterung)
    :param grafik: zu speichernde Grafik
    :param debug: kann optional auf `True` gesetzt werden, um über der Grafik
                  Debug-Informationen darzustellen
    """
    save_graphic(datei_name, grafik, debug)


def speichere_animation(
    datei_name: str, grafiken: list[Grafik], dauer: int = 40, loop: bool = True
):
    """
    Speichere eine Sequenz von Grafiken als Animation (GIF).

    Beim Anzeigen des GIFs werden die Grafiken in einer unendlichen Schleife
    animiert (normalerweise mit 25 Grafiken pro Sekunde).

    :param datei_name: Name der zu kreierenden Datei (mit Erweiterung '.gif')
    :param grafiken: Liste der zu speichernden Grafiken
    :param dauer: Dauer jeder Grafik, in Millisekunden (Default: 40
           Millisekunden, enspricht 25 Grafiken pro Sekunde)
    :param loop: bestimmt ob das GIF in einer unendlichen Schleife abgespielt
           werden soll (Default: true)
    """
    save_animation(datei_name, grafiken, dauer, loop)


def zeige_animation(grafiken: list[Grafik], dauer: int = 40, loop: bool = True):
    """
    Zeige eine Sequenz von Grafiken als Animation (GIF) an.

    Beim Anzeigen des GIFs werden die Grafiken in einer unendlichen Schleife
    animiert (normalerweise mit 25 Grafiken pro Sekunde).

    :param grafiken: Liste der anzuzeigenden Grafiken
    :param dauer: Dauer jeder Grafik, in Millisekunden (Default: 40
           Millisekunden, enspricht 25 Grafiken pro Sekunde)
    :param loop: bestimmt ob das GIF in einer unendlichen Schleife abgespielt
           werden soll (Default: true)
    """
    show_animation(grafiken, dauer, loop)



"""
(Re-)Export names for:
- basic types (Graphic and Color)
- color names
- operations
- primitves
- I/O
from this top-level module.
"""


from pytamaro.color import *
from pytamaro.color_functions import *
from pytamaro.color_names import *
from pytamaro.graphic import *
from pytamaro.io import *
from pytamaro.operations import *
from pytamaro.point import *
from pytamaro.point_names import *
from pytamaro.primitives import *

LANGUAGE = "en"

__version__ = "0.5.0"
